import os

import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler

from bdpn.bd_model import RHO, PSI
from bdpn.bdpn_model import UPSILON, PHI
from bdpn.dl import MODEL2TARGET_COLUMNS, BDCT, TRAINING_PATH, BD
from bdpn.dl import MODEL_PATH, QUANTILES, MODELS, MODEL_FINDER
from bdpn.dl.bdct_model_finder import MODEL_FINDER_PATH
from bdpn.dl.model_serializer import load_scaler_joblib, load_model_keras
from bdpn.dl.training import get_test_data
from bdpn.dl.tree_encoder import forest2sumstat_df, scale_back_array
from bdpn.tree_manager import read_forest



def estimate_cis(forest_sumstats, model_name, Y_pred):
    target_columns = list(MODEL2TARGET_COLUMNS[model_name])
    train_sumstats = pd.read_csv(os.path.join(TRAINING_PATH, f'{model_name}.csv.xz'))[target_columns + [RHO, 'n_trees', 'n_tips']]
    train_results = pd.read_csv(os.path.join(TRAINING_PATH, f'{model_name}.estimates_{model_name}.csv.xz'), index_col=0)[target_columns]
    Y_pred = Y_pred[target_columns]

    scaler_y = StandardScaler()
    train_results = scaler_y.fit_transform(train_results.to_numpy(dtype=float, na_value=0))
    Y_pred = scaler_y.transform(Y_pred.to_numpy(dtype=float, na_value=0))

    Y_pred = pd.DataFrame(Y_pred, columns=target_columns, index=forest_sumstats.index)
    train_results = pd.DataFrame(train_results, columns=target_columns)

    n_train_examples = len(train_sumstats)
    tip_threshold = 200_000 if n_train_examples > 200_000 else 40_000
    tree_threshold = 100_000 if n_train_examples > 200_000 else 20_000
    rho_threshold = 10_000
    param_threshold = 1000


    for index in forest_sumstats.index:
        rho, n_trees, n_tips = forest_sumstats.loc[index, [RHO, 'n_trees', 'n_tips']]
        cur_sumstats = train_sumstats
        cur_results = train_results
        # print('tips', n_tips, np.percentile(train_sumstats['n_tips'], [0, 5, 50, 95, 100]))
        cur_ids = cur_sumstats.iloc[np.abs(cur_sumstats[RHO] - rho).argsort(), :].index[:tip_threshold]
        cur_sumstats = cur_sumstats.loc[cur_ids, :]
        cur_results = cur_results.loc[cur_ids, :]
        # print('tips', n_tips, np.percentile(cur_sumstats['n_tips'], [0, 5, 50, 95, 100]))
        # print('trees', n_trees, np.percentile(cur_sumstats['n_trees'], [0, 5, 50, 95, 100]))
        cur_ids = cur_sumstats.iloc[np.abs(cur_sumstats['n_trees'] - n_trees).argsort(), :].index[:tree_threshold]
        cur_sumstats = cur_sumstats.loc[cur_ids, :]
        cur_results = cur_results.loc[cur_ids, :]
        # print('trees', n_trees, np.percentile(cur_sumstats['n_trees'], [0, 5, 50, 95, 100]))
        # print(RHO, rho, np.percentile(cur_sumstats[RHO], [0, 5, 50, 95, 100]))
        cur_ids = cur_sumstats.iloc[np.abs(cur_sumstats['n_tips'] - n_tips).argsort(), :].index[:rho_threshold]
        cur_sumstats = cur_sumstats.loc[cur_ids, :]
        cur_results = cur_results.loc[cur_ids, :]
        # print(RHO, rho, np.percentile(cur_sumstats[RHO], [0, 5, 50, 95, 100]))

        loss = np.abs(cur_sumstats[target_columns] - Y_pred.loc[index, target_columns]).sum(axis=1)
        cur_ids = cur_sumstats.iloc[loss.argsort(), :].index[:param_threshold]
        cur_sumstats = cur_sumstats.loc[cur_ids, :]
        cur_results = cur_results.loc[cur_ids, :]


        for col in target_columns:
            val = Y_pred.loc[index, col]
            # # print(col, val, np.percentile(cur_sumstats[col], [0, 5, 50, 95, 100]), np.percentile(cur_results[col], [0, 5, 50, 95, 100]))
            # par_ids = cur_sumstats.iloc[np.abs(cur_sumstats[col] - val).argsort(), :].index[:param_threshold]
            # true_vals = cur_sumstats.loc[par_ids, col]
            # pred_vals = cur_results.loc[par_ids, col]
            # # print(col, val, np.percentile(true_vals, [0, 5, 50, 95, 100]), np.percentile(pred_vals, [0, 5, 50, 95, 100]))
            # errors = pred_vals - true_vals
            errors = cur_results[col] - cur_sumstats[col]
            dist = val + errors - np.median(errors)
            # dist = np.maximum(dist, 0)
            # if UPSILON == col:
            #     dist = np.minimum(dist, 1)
            Y_pred.loc[index, f'{col}_2.5'] = np.percentile(dist, 2.5)
            Y_pred.loc[index, f'{col}_97.5'] = np.percentile(dist, 97.5)
    for suffix in ('', '_2.5', '_97.5'):
        Y_pred[[f'{col}{suffix}' for col in target_columns]] = \
            scaler_y.inverse_transform(Y_pred[[f'{col}{suffix}' for col in target_columns]].to_numpy(dtype=float, na_value=0))


    Y_pred = np.maximum(Y_pred, 0)
    ups_cols = [col for col in Y_pred.columns if UPSILON in col]
    if ups_cols:
        Y_pred[ups_cols] = np.minimum(Y_pred[ups_cols], 1)
    return Y_pred


def predict_parameters(forest_sumstats, model_name=MODEL_FINDER, ci=False):
    models_ext = MODELS + (BDCT,)
    n_forests = len(forest_sumstats)

    if MODEL_FINDER == model_name:
        from bdpn.dl import training_model_finder

        X = training_model_finder.get_test_data(df=forest_sumstats)
        model_weights = np.concat([load_model_keras(MODEL_FINDER_PATH).predict(X), np.zeros((n_forests, 1), dtype=float)], axis=1)
    else:
        model_weights = np.zeros((n_forests, len(models_ext)))
        model_weights[:, models_ext.index(model_name)] = 1



    X, SF = get_test_data(dfs=[forest_sumstats], scaler_x=None)

    results = []

    # result = pd.DataFrame(index=np.arange(X.shape[0]))

    model_ids = [i for i in range(len(models_ext)) if not np.all(model_weights[:, i] == 0)]
    for model_id in model_ids:
        model_name = models_ext[model_id]

        model_path = os.path.join(MODEL_PATH, model_name)

        X_cur, SF_cur = np.array(X), np.array(SF)

        scaler_x = load_scaler_joblib(model_path, suffix='x')
        if scaler_x:
            X_cur = scaler_x.transform(X_cur)

        Y_pred = load_model_keras(model_path).predict(X_cur)
        scaler_y = load_scaler_joblib(model_path, suffix='y')
        n_quant = len(QUANTILES)
        target_columns = MODEL2TARGET_COLUMNS[model_name]
        for i in range(n_quant):
            Y_pred[:, i::n_quant] = scaler_y.inverse_transform(Y_pred[:, i::n_quant])

        tc_cis = tuple((f'{_}_{q * 100:.1f}' if q != 0.5 else _) for _ in target_columns
                       for q in QUANTILES)
        Y_pred = np.maximum(Y_pred, 0)
        scale_back_array(Y_pred, SF_cur, tc_cis)
        idx = [i for i in range(len(tc_cis)) if tc_cis[i].startswith(UPSILON)]
        Y_pred[:, idx] = np.minimum(Y_pred[:, idx], 1)
        results.append(pd.DataFrame(Y_pred, columns=tc_cis))

    if len(model_ids) == 1:
        result = results[0]
    else:
        BD_idx = models_ext.index(BD)
        if BD_idx in model_ids:
            ups_cols = [(f'{UPSILON}_{q * 100:.1f}' if q != 0.5 else UPSILON) for q in QUANTILES]
            results[BD_idx].loc[:, ups_cols] = np.zeros((n_forests, len(ups_cols)), dtype=float)
            phi_cols = [(f'{PHI}_{q * 100:.1f}' if q != 0.5 else PHI) for q in QUANTILES]
            psi_cols = [(f'{PSI}_{q * 100:.1f}' if q != 0.5 else PSI) for q in QUANTILES]
            results[BD_idx].loc[:, phi_cols] = results[BD_idx].loc[:, psi_cols].to_numpy(dtype=float, na_value=0)
        columns = results[0].columns
        result = pd.DataFrame(index=forest_sumstats.index)
        for col in columns:


            predictions = np.array([res[col].to_numpy(dtype=float, na_value=0) for res in results]).T
            weights = model_weights[:, model_ids]
            result[col] = np.average(predictions, weights=weights, axis=1)

    return result


def main():
    """
    Entry point for tree parameter estimation with a BDCT model with command-line arguments.
    :return: void
    """
    import argparse

    parser = \
        argparse.ArgumentParser(description="Estimate BDCT parameters.")
    parser.add_argument('--model_name', choices=MODELS + (MODEL_FINDER, BDCT), default=MODEL_FINDER, type=str,
                        help=f'BDCT model flavour. If {MODEL_FINDER} is specified, '
                             f'model finder will be used to pick the model.')
    parser.add_argument('--p', default=0, type=float, help='sampling probability')
    parser.add_argument('--log', required=True, type=str, help="output log file")
    parser.add_argument('--nwk', default=None, type=str, help="input tree file")
    parser.add_argument('--sumstats', default=None, type=str, help="input tree file(s) encoded as sumstats")
    parser.add_argument('--ci', action='store_true', help="calculate CIs")
    params = parser.parse_args()

    if not params.sumstats:
        if params.p <= 0 or params.p > 1:
            raise ValueError('The sampling probability must be grater than 0 and not greater than 1.')

        forest = read_forest(params.nwk)
        print(f'Read a forest of {len(forest)} trees with {sum(len(_) for _ in forest)} tips in total')
        forest_df = forest2sumstat_df(forest, rho=params.p)
    else:
        forest_df = pd.read_csv(params.sumstats)
    predict_parameters(forest_df, model_name=params.model_name, ci=params.ci).to_csv(params.log, header=True)


if '__main__' == __name__:
    main()